""" A module mimicking some functionality of uflacs' modified terminals """

from ufl.algorithms import MultiFunction
from dune.perftool import Restriction


class ModifiedTerminalTracker(MultiFunction):
    """ A multifunction base that defines handler for
    grad, reference_grad, positive_restricted and negative_restricted.
    The appearance of those classes changes the internal state of the MF.
    """
    def __init__(self):
        MultiFunction.__init__(self)
        self.grad = False
        self.reference_grad = False
        self.restriction = Restriction.NONE

    def positive_restricted(self, o):
        assert self.restriction == Restriction.NONE
        self.restriction = Restriction.POSITIVE

        from dune.perftool.generation import global_context
        with global_context(restriction=Restriction.POSITIVE):
            ret = self.call(o.ufl_operands[0])

        self.restriction = Restriction.NONE
        return ret

    def negative_restricted(self, o):
        assert self.restriction == Restriction.NONE
        self.restriction = Restriction.NEGATIVE

        from dune.perftool.generation import global_context
        with global_context(restriction=Restriction.NEGATIVE):
            ret = self.call(o.ufl_operands[0])

        self.restriction = Restriction.NONE
        return ret

    def grad(self, o):
        assert not self.grad and not self.reference_grad
        self.grad = True
        ret = self.call(o.ufl_operands[0])
        self.grad = False
        return ret

    def reference_grad(self, o):
        assert not self.reference_grad and not self.grad
        self.reference_grad = True
        ret = self.call(o.ufl_operands[0])
        self.reference_grad = False
        return ret


class ModifiedArgumentDescriptor(MultiFunction):
    def __init__(self, e):
        MultiFunction.__init__(self)

        self.grad = False
        self.reference_grad = False
        self.index = None
        self.restriction = Restriction.NONE
        self.expr = e

        self.__call__(e)
        self.__call__ = None

    def __eq__(self, other):
        return self.expr == other.expr

    def grad(self, o):
        self.grad = True
        self(o.ufl_operands[0])

    def reference_grad(self, o):
        self.reference_grad = True
        self(o.ufl_operands[0])

    def positive_restricted(self, o):
        self.restriction = Restriction.POSITIVE
        self(o.ufl_operands[0])

    def negative_restricted(self, o):
        self.restriction = Restriction.NEGATIVE
        self(o.ufl_operands[0])

    def indexed(self, o):
        self.index = o.ufl_operands[1]
        self(o.ufl_operands[0])

    def argument(self, o):
        self.argexpr = o

    def coefficient(self, o):
        self.argexpr = o


class _ModifiedArgumentExtractor(MultiFunction):
    """ A multifunction that extracts and returns the set of modified arguments """

    def __call__(self, o, argnumber=None, testfunction=True, trialfunction=False):
        self.argnumber = argnumber
        self.trialfunction = trialfunction
        self.testfunction = testfunction
        self.modified_arguments = set()
        ret = self.call(o)
        if ret:
            # This indicates that this entire expression was a modified thing...
            self.modified_arguments.add(ret)
        return tuple(ModifiedArgumentDescriptor(ma) for ma in self.modified_arguments)

    def expr(self, o):
        for op in o.ufl_operands:
            ret = self.call(op)
            if ret:
                self.modified_arguments.add(ret)

    def indexed(self, o):
        if self.call(o.ufl_operands[0]):
            return o

    def positive_restricted(self, o):
        if self.call(o.ufl_operands[0]):
            return o

    def negative_restricted(self, o):
        if self.call(o.ufl_operands[0]):
            return o

    def grad(self, o):
        if self.call(o.ufl_operands[0]):
            return o

    def reference_grad(self, o):
        if self.call(o.ufl_operands[0]):
            return o

    def argument(self, o):
        if self.testfunction:
            if self.argnumber is None or o.number() == self.argnumber:
                return o

    def coefficient(self, o):
        if self.trialfunction:
            if o.count() == 0:
                return o

    call = MultiFunction.__call__


def extract_modified_arguments(expr, **kwargs):
    return _ModifiedArgumentExtractor()(expr, **kwargs)


class _ModifiedArgumentNumber(MultiFunction):
    """ return the number() of a modified argument """
    def expr(self, o):
        return self(o.ufl_operands[0])

    def argument(self, o):
        return o.number()


def modified_argument_number(expr):
    """ Given an expression, return the number() of the argument in it """
    return _ModifiedArgumentNumber()(expr)
